import os
import pickle
import gzip
import traceback
import pytz
import tempfile
import contextlib
import time
import pandas as pd
import datatable
from qtc.ext.enum import Enum
import qtc.utils.datetime_utils as dtu
import qtc.utils.misc_utils as mu
import qtc.utils.file_system_utils as fsu
from qtc.ext.logging import set_logger
logger = set_logger()


class DateTimeFolderStructureMode(Enum):
    NON_DATETIME = 'NON_DATETIME'
    DATEID = 'DateId'
    YEAR_MONTH = '%Y/%Y%m'
    YEAR_MONTH_DATEID = '%Y/%Y%m/%Y%m%d'
    YEAR_DATEID = '%Y/%Y%m%d'
    YEAR = '%Y'
    YEAR_MONTH_DATESTR_HYPHEN_SEP = '%Y/%Y-%m/%Y-%m-%d'
    DATEID_HYPHEN_TIMESTR = '%Y%m%d-%H%M%S'
    DATETIME_TIMEZONE = '%Y%m%dT%H%M%S%z'
    YEAR_MONTH_DATESTR_HYPHEN_SEP_DATEID_HYPHEN_TIMESTR = '%Y/%Y-%m/%Y-%m-%d/%Y%m%d-%H%M%S'
    CURRENT = 'current'


class DateTimeFoldersBuilder:
    @property
    def root(self):
        return self.__root

    # @root.setter
    # def root(self, root):
    #     self.__root = root
    # #

    @property
    def datetime_folder_structure_mode(self):
        return self.__datetime_folder_structure_mode

    # @datetime_folder_structure_mode.setter
    # def datetime_folder_structure_mode(self, datetime_folder_structure_mode):
    #     self.__datetime_folder_structure_mode = datetime_folder_structure_mode
    # #

    @property
    def timezone(self):
        return self.__timezone

    # @timezone.setter
    # def timezone(self, timezone):
    #     self.__timezone = timezone
    # #

    def __init__(self, root, datetime_folder_structure_mode, timezone=None):
        self.__root = root
        if isinstance(datetime_folder_structure_mode, str):
            datetime_folder_structure_mode = DateTimeFolderStructureMode.from_name(datetime_folder_structure_mode)

        self.__datetime_folder_structure_mode = datetime_folder_structure_mode

        if timezone is None:
            timezone = pytz.timezone('UTC')

        self.__timezone = pytz.timezone(timezone) if isinstance(timezone, str) else timezone

    def _get_dt_folder_structure(self, dt, subfolders=None):
        """
        >>> from qtc.file.file_serialization import DateTimeFoldersBuilder
        >>> dfm1 = DateTimeFoldersBuilder(root='/bam/aws/devel/quant_risk/dw/positions',
                                          datetime_folder_structure_mode='DATEID',
                                          timezone='America/New_York')
        >>> dfm1._get_dt_folder_structure(dt='20200806-094916')
        '/bam/aws/devel/quant_risk/dw/positions/20200806'
        >>> dfm2 = DateTimeFoldersBuilder(root='/mnt/rsk/Risk/EPA_Hist/POMO-ALL_EQUITY_ACUS',
                                          datetime_folder_structure_mode='YEAR_MONTH_DAY',
                                          timezone='UTC')
        >>> dfm2._get_dt_folder_structure(dt='20200806-094916')
        '/mnt/rsk/Risk/EPA_Hist/POMO-ALL_EQUITY_ACUS/2020/202008/20200806'
        >>> fss = DateTimeFoldersBuilder(root='/bam/aws/devel/quant_risk/dw/positions',
                                         datetime_folder_structure_mode='YEAR_MONTH_DATESTR_HYPHEN_SEP',
                                         timezone='UTC')
        >>> fss._get_dt_folder_structure(dt='20200806-094916')
        '/bam/aws/devel/quant_risk/dw/positions/2020/2020-08/2020-08-06'
        >>> fss = DateTimeFoldersBuilder(root='/bam/aws/devel/quant_risk/dw/positions',
                                         datetime_folder_structure_mode='DATETIME',
                                         timezone='UTC')
        '/bam/aws/devel/quant_risk/dw/positions/20200806-094916'
        >>> dfm3 = DateTimeFoldersBuilder(root='/mnt/rsk/Risk/ESQA/Intraday',
                                          datetime_folder_structure_mode='YEAR_MONTH_DATESTR_HYPHEN_SEP_DATEID_HYPHEN_TIMESTR',
                                          timezone='UTC')
        >>> dfm3._get_dt_folder_structure(dt='20220222-165800')
        '/mnt/rsk/Risk/ESQA/Intraday/2022/2022-02/2022-02-22/20220222-165800'
        """

        parent_folder = self.root
        if subfolders is not None:
            parent_folder = os.path.join(parent_folder, *mu.iterable_to_tuple(subfolders, raw_type='str'))

        dt = dtu.normalize_dt(dt=dt, timezone=self.timezone)

        if self.datetime_folder_structure_mode == DateTimeFolderStructureMode.NON_DATETIME:
            folder = parent_folder
        elif self.datetime_folder_structure_mode == DateTimeFolderStructureMode.DATEID:
            folder = os.path.join(parent_folder, dt.strftime('%Y%m%d'))
        elif self.datetime_folder_structure_mode == DateTimeFolderStructureMode.YEAR_MONTH:
            folder = os.path.join(parent_folder, dt.strftime('%Y'), dt.strftime('%Y%m'))
        elif self.datetime_folder_structure_mode == DateTimeFolderStructureMode.YEAR_MONTH_DATEID:
            folder = os.path.join(parent_folder, dt.strftime('%Y'), dt.strftime('%Y%m'), dt.strftime('%Y%m%d'))
        elif self.datetime_folder_structure_mode == DateTimeFolderStructureMode.YEAR_DATEID:
            folder = os.path.join(parent_folder, dt.strftime('%Y'), dt.strftime('%Y%m%d'))
        elif self.datetime_folder_structure_mode == DateTimeFolderStructureMode.YEAR:
            folder = os.path.join(parent_folder, dt.strftime('%Y'))
        elif self.datetime_folder_structure_mode == DateTimeFolderStructureMode.YEAR_MONTH_DATESTR_HYPHEN_SEP:
            folder = os.path.join(parent_folder, dt.strftime('%Y'), dt.strftime('%Y-%m'), dt.strftime('%Y-%m-%d'))
        elif self.datetime_folder_structure_mode == DateTimeFolderStructureMode.YEAR_MONTH_DATESTR_HYPHEN_SEP_DATEID_HYPHEN_TIMESTR:
            folder = os.path.join(parent_folder, dt.strftime('%Y'), dt.strftime('%Y-%m'), dt.strftime('%Y-%m-%d'), dt.strftime('%Y%m%d-%H%M%S'))
        elif self.datetime_folder_structure_mode == DateTimeFolderStructureMode.DATEID_HYPHEN_TIMESTR:
            folder = os.path.join(parent_folder, dt.strftime('%Y%m%d-%H%M%S'))
        elif self.datetime_folder_structure_mode == DateTimeFolderStructureMode.CURRENT:
            folder = os.path.join(parent_folder, 'current')

        return folder

    def compile_folder(self, dt=None, subfolders=None, is_subfolders_under_dt_folder=True):
        """
        >>> from qtc.file.file_serialization import DateTimeFoldersBuilder
        >>> fs = DateTimeFoldersBuilder(root='/bam/aws/devel/quant_risk/dw/positions',
                                        datetime_folder_structure_mode='DATEID',
                                        timezone='America/New_York')
        >>> from qtc.file.file_serialization import FileSerializer
        >>> fsd = FileSerializer(filename_prefix='factor_loadings',
                                 filename_dt_component_type='DATEID')
        >>> fs.compile_folder(file_serializer=fsd,
                              dt=20210929,
                              extra_filename_components='Archive,BARRA-USFASTD')
        """

        dt = dtu.normalize_dt(dt=dt, timezone=self.timezone)

        if is_subfolders_under_dt_folder:
            folder = self._get_dt_folder_structure(dt=dt)
            if subfolders is not None:
                folder = os.path.join(folder, *mu.iterable_to_tuple(subfolders, raw_type='str'))
        else:
            folder = self._get_dt_folder_structure(dt=dt, subfolders=subfolders)

        return folder

    def __repr__(self):
        return f"{type(self).__name__}(" \
               f"root='{self.root}', " \
               f"datetime_folder_structure_mode={self.datetime_folder_structure_mode}, " \
               f"timezone={self.timezone})"


class FileSerializationProtocol(Enum):
    TXT = 'txt'
    CSV = 'csv'
    PKL = 'pkl'
    PKL_GZ = 'pkl.gz'
    PICKLE = 'pickle'
    PICKLE_GZ = 'pickle.gz'
    LOG = 'log'
    HTML = 'html'
    FEATHER = 'feather'
    FTR = 'ftr'
    JAY = 'jay'
    JAY_GZ = 'jay.gz'
    PDF = 'pdf'

    @staticmethod
    def suffix(protocol):
        try:
            if isinstance(protocol, str):
                protocol = FileSerializationProtocol.retrieve(protocol)

            ret = protocol.value
        except:
            logger.warn(f'Faild to retrieve FileSerializationProtocol with value={protocol} !')
            ret = protocol

        return ret

    @staticmethod
    def protocol(suffix):
        if suffix.startswith('.'):
            suffix = suffix[1:]

        try:
            return FileSerializationProtocol(suffix.lower())
        except Exception as e:
            logger.warn(f"suffix={suffix} is not supported in [txt|csv|pkl|pkl.gz|html|feather|ftr']. Use pkl.gz by default! ")
            return FileSerializationProtocol.PICKLE_GZ


class FilenameDateTimeComponentType(Enum):
    NON_DATETIME = 'NON_DATETIME'
    DATEID = 'DateId'
    DATEID_HYPHEN_TIMESTR = '%Y%m%d-%H%M%S'
    DATEID_T_TIMESTR_OFFSETSTR = '%Y%m%dT%H%M%S%z'


@contextlib.contextmanager
def atomic_overwrite(file_path, folder=None, text=True, gzip_on=False):
    if folder is None:
        folder = os.path.dirname(file_path)

    if gzip_on:
        raise Exception(f'Not ready yet for gzip_on={gzip_on}!')
        # # fd, tmp_path = tempfile.mkstemp(dir=folder, text=text)
        # if text:
        #     fd = tempfile.NamedTemporaryFile(mode='w', dir=folder, delete=False)
        #     tmp_path = fd.name
        #     with gzip.open(tmp_path, mode='w') as f:
        #         yield f
        # else:
        #     fd = tempfile.NamedTemporaryFile(mode='w+b', dir=folder, delete=False)
        #     tmp_path = fd.name
        #     with gzip.GzipFile(filename=tmp_path, mode='wb') as f:
        #         yield f
    else:
        fd, tmp_path = tempfile.mkstemp(dir=folder, text=text)
        if text:
            with os.fdopen(fd, 'w', encoding='utf-8') as f:
                yield f
        else:
            with os.fdopen(fd, 'wb') as f:
                yield f

    error_count = 0
    while True:
        try:
            os.replace(tmp_path, file_path)  # this will only happen if no exception was raised
            break
        except PermissionError as e:
            error_count += 1  # Error with replacing, likely being replaced by another process

            if (error_count >= 10):
                raise

            logger.debug(f'Failed writing {file_path} - waiting {error_count} seconds')
            time.sleep(1 * error_count)  # Sleep increasing amounts of time

    if error_count > 0:
        logger.warn(f'Failed writing {file_path} {error_count} times')


class FileSerializer:
    @property
    def filename_prefix(self):
        return self.__filename_prefix

    @property
    def filename_exclude_str(self):
        return self.__filename_exclude_str

    @property
    def filename_dt_component_type(self):
        return self.__filename_dt_component_type

    @property
    def protocol(self):
        return self.__protocol

    @property
    def suffix(self):
        return self.__suffix

    @property
    def filename_component_sep(self):
        return self.__filename_component_sep

    @property
    def kwargs(self):
        return self.__kwargs

    def __init__(self, filename_prefix,
                 filename_dt_component_type,
                 protocol='PKL.GZ',
                 filename_component_sep='.',
                 filename_exclude_str=None,
                 **kwargs):
        self.__filename_prefix = filename_prefix
        self.__filename_exclude_str = filename_exclude_str

        if isinstance(filename_dt_component_type, str):
            filename_dt_component_type = FilenameDateTimeComponentType.retrieve(filename_dt_component_type)

        self.__filename_dt_component_type = filename_dt_component_type

        if isinstance(protocol, str):
            protocol = FileSerializationProtocol.retrieve(protocol)

        self.__protocol = protocol

        self.__suffix = FileSerializationProtocol.suffix(self.protocol)
        self.__filename_component_sep = filename_component_sep
        self.__kwargs = kwargs

    def _get_file_dt_component(self, dt):
        if self.filename_dt_component_type == FilenameDateTimeComponentType.NON_DATETIME:
            filename_dt_component = None
        elif self.filename_dt_component_type == FilenameDateTimeComponentType.DATEID:
            filename_dt_component = str(dtu.datetime_to_dateid(date=dt))
        elif self.filename_dt_component_type == FilenameDateTimeComponentType.DATEID_HYPHEN_TIMESTR:
            filename_dt_component = dt.strftime('%Y%m%d-%H%M%S')
        elif self.filename_dt_component_type == FilenameDateTimeComponentType.DATEID_T_TIMESTR_OFFSETSTR:
            filename_dt_component = dt.strftime('%Y%m%dT%H%M%S%z')

        return filename_dt_component

    def compile_filename(self, dt=None,
                         extra_filename_components=None,
                         timezone=None, include_run_dt=False):
        """
        >>> from qtc.file.file_serialization import FileSerializer
        >>> fsd = FileSerializer(filename_prefix='factor_loadings',
                                 filename_dt_component_type='DATEID')
        >>> fsd.compile_filename(dt=20210929, extra_filename_components='Archive,AXUS4-MH')
        'factor_loadings.Archive.AXUS4-MH.20210929.pkl.gz'
        """

        dt = dtu.normalize_dt(dt=dt, timezone=timezone)

        if extra_filename_components is None:
            extra_filename_components = list()
        else:
            extra_filename_components = list(mu.iterable_to_tuple(extra_filename_components, raw_type='str'))

        filename_dt_component = self._get_file_dt_component(dt=dt)
        filename_components = [self.filename_prefix] + extra_filename_components
        if filename_dt_component is not None:
            filename_components.append(filename_dt_component)

        filename = self.filename_component_sep.join(filename_components)
        filename = f'{filename}.{self.suffix}'

        return filename

    def search_file(self, folder, dt=None,
                    extra_filename_components=None,
                    timezone=None, allow_as_of=False,
                    log_error=True):
        """
        >>> from qtc.file.file_serialization import FileSerializer
        >>> fsd = FileSerializer(filename_prefix='basket_decomp',
                                 filename_dt_component_type='DATEID',
                                 protocol='PICKLE_GZ')
        >>> from qtc.ext.path import quant_risk_dir
        >>> fsd.search_file(folder=quant_risk_dir('dw','basket_decomp','2021','202109','20210929'))
        '/bam/aws/devel/quant_risk/dw/basket_decomp/2021/202109/20210929/basket_decomp.20210929.pickle.gz'
        """
        if folder is None:
            raise Exception(f'Pleaes provide folder explicitly !')

        if dt is None:
            allow_as_of = True

        dt = dtu.normalize_dt(dt=dt, timezone=timezone)

        if extra_filename_components is None:
            extra_filename_components = list()
        else:
            extra_filename_components = list(mu.iterable_to_tuple(extra_filename_components, raw_type='str'))

        filename_components = [self.filename_prefix] + extra_filename_components
        if not allow_as_of:
            filename_dt_component = self._get_file_dt_component(dt=dt)
            if filename_dt_component is not None:
                filename_components.append(filename_dt_component)

        sep = f'{self.filename_component_sep}*'
        file_pattern = sep.join(filename_components)
        if len(filename_components)==1 and self.filename_dt_component_type!=FilenameDateTimeComponentType.NON_DATETIME:
            file_pattern += sep

        file_pattern = f'{file_pattern}.{self.suffix}'

        files = fsu.search_files_by_pattern(file_patten=file_pattern, folder=folder,
                                            exclude_str=self.filename_exclude_str)

        if allow_as_of:
            if self.filename_dt_component_type!=FilenameDateTimeComponentType.NON_DATETIME:
                files = [f for f in files if fsu.is_earlier(path=f.replace(f'.{self.suffix}',''),
                                                            target_dt=dt,
                                                            filename_sep=self.filename_component_sep, timezone=timezone)]
            if len(files)>=1:
                return sorted(files)[-1]
            else:
                if log_error:
                    logger.error(f'No file with pattern {file_pattern} as of {dt} found in {folder}!')

                return None
        else:
            if len(files)==1:
                return files[0]

            if len(files)==0:
                if log_error:
                    logger.error(f'Expecting exactly ONE file with pattern {file_pattern}. '
                                 f'However {len(files)} files found in {folder} !')
                return None

            if log_error:
                logger.warn(f'Expecting exactly ONE file with pattern {file_pattern}. '
                            f'Use the first one in the following {len(files)} files found in {folder} !\n{files}')

            return files[0]

    def _serialize(self, data, file_path):
        if data is None:
            logger.warn(f"data is None. Skip serializing '{self.filename_prefix}' -> {file_path}")
            return None

        dirname = os.path.dirname(file_path)
        if dirname is not None and dirname.strip()!='':
            os.makedirs(dirname, exist_ok=True)

        try:
            if self.protocol == FileSerializationProtocol.TXT or \
               self.protocol == FileSerializationProtocol.LOG or \
               self.protocol == FileSerializationProtocol.HTML:
                with atomic_overwrite(file_path=file_path, text=True, gzip_on=False) as fh:
                    fh.write(data)
            elif self.protocol == FileSerializationProtocol.CSV:
                with atomic_overwrite(file_path=file_path, text=True, gzip_on=False) as fh:
                    if isinstance(data, pd.DataFrame) or isinstance(data, pd.Series):
                        data.to_csv(fh, **self.kwargs)
                    else:
                        fh.write(data)
            elif self.protocol == FileSerializationProtocol.PICKLE or \
                 self.protocol == FileSerializationProtocol.PKL:
                with atomic_overwrite(file_path=file_path, text=False, gzip_on=False) as fh:
                    pickle.dump(data, fh)
            elif self.protocol == FileSerializationProtocol.PICKLE_GZ or \
                 self.protocol == FileSerializationProtocol.PKL_GZ:
                # with atomic_overwrite(file_path=file_path, text=False, gzip_on=True) as fh:
                #     # pickle.dump(data, fh)
                #     temp = pickle.dumps(data)
                #     print(temp)
                #     fh.write(temp)
                # #
                with gzip.open(file_path, 'wb') as fh:
                    pickle.dump(data, fh)
            elif self.protocol == FileSerializationProtocol.FEATHER:
                with atomic_overwrite(file_path=file_path, text=False, gzip_on=False) as fh:
                    if not isinstance(data, pd.DataFrame) and not isinstance(data, pd.Series):
                        raise Exception(f'type(data)={type(data)} is not supported with protocol={self.protocol} !')

                    data.to_feather(fh, compression='uncompressed', **self.kwargs)
            elif self.protocol == FileSerializationProtocol.FTR:
                with atomic_overwrite(file_path=file_path, text=False, gzip_on=False) as fh:
                    if not isinstance(data, pd.DataFrame) and not isinstance(data, pd.Series):
                        raise Exception(f'type(data)={type(data)} is not supported with protocol={self.protocol} !')

                    data.to_feather(fh, compression='lz4', **self.kwargs)
            elif self.protocol == FileSerializationProtocol.JAY:
                with atomic_overwrite(file_path=file_path, text=False, gzip_on=False) as fh:
                    if not isinstance(data, pd.DataFrame) and not isinstance(data, pd.Series):
                        raise Exception(f'type(data)={type(data)} is not supported with protocol={self.protocol} !')

                    for col in data.columns:
                        if isinstance(data[col].dtype, pd.Int64Dtype):
                            data[col] = data[col].astype(float)

                    fh.write(datatable.Frame(data).to_jay())
            elif self.protocol == FileSerializationProtocol.JAY_GZ:
                with gzip.open(file_path, 'wb') as fh:
                    fh.write(datatable.Frame(data).to_jay())
            else:
                raise Exception(f'Serialization for protocol={self.protocol} not supported yet !')

            if isinstance(data, pd.DataFrame) or isinstance(data, pd.Series):
                logger.info(f"'{self.filename_prefix}' with shape {data.shape} serialized -> {file_path}")
            elif isinstance(data, dict):
                logger.info(f"'{self.filename_prefix}' with keys {data.keys()} serialized -> {file_path}")
            else:
                logger.info(f"'{self.filename_prefix}' serialized -> {file_path}")

            return file_path
        except Exception as e:
            logger.error(f"Failed to serialize '{self.filename_prefix}' -> {file_path}")
            traceback.print_exc()
            return None

    def serialize(self, data,
                  folder=None,
                  dt=None,
                  extra_filename_components=None,
                  timezone=None,
                  include_run_dt=False):
        filename = self.compile_filename(dt=dt,
                                         extra_filename_components=extra_filename_components,
                                         timezone=timezone,
                                         include_run_dt=include_run_dt)
        file_path = filename if folder is None else os.path.join(folder, filename)

        return self._serialize(data=data, file_path=file_path)

    def _deserialize(self, file_path, log_error=False):
        data = None
        try:
            if self.protocol == FileSerializationProtocol.TXT or \
               self.protocol == FileSerializationProtocol.LOG or \
               self.protocol == FileSerializationProtocol.HTML:
                with open(file_path, 'rt') as fh:
                    data = fh.read()
            elif self.protocol == FileSerializationProtocol.CSV:
                kwargs = self.kwargs.copy()
                kwargs.pop('index', None)
                data = pd.read_csv(file_path, **kwargs)
            elif self.protocol == FileSerializationProtocol.PICKLE or \
                 self.protocol == FileSerializationProtocol.PKL:
                try:
                    data = pd.read_pickle(file_path)
                except:
                    with open(file_path, 'rb') as fh:
                        data = pickle.load(fh)
            elif self.protocol == FileSerializationProtocol.PICKLE_GZ or \
                 self.protocol == FileSerializationProtocol.PKL_GZ:
                try:
                    data = pd.read_pickle(file_path)
                except:
                    with gzip.open(file_path, 'rb') as fh:
                        data = pickle.load(fh)
            elif self.protocol == FileSerializationProtocol.FEATHER or \
                 self.protocol == FileSerializationProtocol.FTR:
                data = pd.read_feather(file_path)
            elif self.protocol == FileSerializationProtocol.JAY:
                data = datatable.fread(file_path).to_pandas()
            elif self.protocol == FileSerializationProtocol.JAY_GZ:
                with gzip.open(file_path, 'rb') as fh:
                    content = fh.read()
                data = datatable.fread(content).to_pandas()
            if isinstance(data, pd.DataFrame) or isinstance(data, pd.Series):
                logger.info(f"'{self.filename_prefix}' with shape {data.shape} deserialized <- {file_path}")
            else:
                logger.info(f"'{self.filename_prefix}' deserialized <- {file_path}")
        except Exception as e:
            if log_error:
                logger.error(f"Failed to deserialize <- {file_path}")
                traceback.print_exc()

        return data

    def deserialize(self,
                    folder=None,
                    dt=None,
                    extra_filename_components=None,
                    timezone=None, allow_as_of=False,
                    log_error=True):
        file_path = self.search_file(folder=folder, dt=dt,
                                     extra_filename_components=extra_filename_components,
                                     timezone=timezone, allow_as_of=allow_as_of,
                                     log_error=log_error)
        return self._deserialize(file_path=file_path, log_error=log_error)

    def __repr__(self):
        return f"{type(self).__name__}(filename_prefix='{self.filename_prefix}', " \
               f"protocol={self.protocol}, " \
               f"filename_dt_component_type={self.filename_dt_component_type}, " \
               f"filename_component_sep='{self.filename_component_sep}')"


class DateTimeFoldersBuilderManager:
    __registry = dict()

    @classmethod
    def get(cls, root, datetime_folder_structure_mode, timezone=None):
        if root is None:
            raise Exception(f'Please provide root explicitly !')

        if isinstance(datetime_folder_structure_mode, str):
            datetime_folder_structure_mode = DateTimeFolderStructureMode.retrieve(datetime_folder_structure_mode)

        key = (root, datetime_folder_structure_mode, timezone)
        if key not in DateTimeFoldersBuilderManager.__registry:
            DateTimeFoldersBuilderManager.__registry[key] = DateTimeFoldersBuilder(root=root,
                                                                                   datetime_folder_structure_mode=datetime_folder_structure_mode,
                                                                                   timezone=timezone)
        # else:
        #     logger.warn(f'Skipping creating a new DateTimeFoldersHelper object since '
        #                 f'key=(root={root}, datetime_folder_structure_mode={datetime_folder_structure_mode}, timezone={timezone}) '
        #                 f'has been registered!')
        #
        return DateTimeFoldersBuilderManager.__registry[key]


class FileSerializerManager:
    __registry = dict()
    __named_registry = dict()

    @classmethod
    def get(cls, filename_prefix, filename_dt_component_type,
            protocol='PKL.GZ',
            filename_component_sep='.',
            filename_exclude_str=None,
            **kwargs):
        if isinstance(filename_dt_component_type, str):
            filename_dt_component_type = FilenameDateTimeComponentType.retrieve(filename_dt_component_type)

        if isinstance(protocol, str):
            protocol = FileSerializationProtocol.retrieve(protocol)

        key = (filename_prefix, filename_dt_component_type, protocol, filename_component_sep, filename_exclude_str)
        if key not in FileSerializerManager.__registry:
            FileSerializerManager.__registry[key] = \
                FileSerializer(filename_prefix=filename_prefix,
                               filename_dt_component_type=filename_dt_component_type,
                               protocol=protocol,
                               filename_component_sep=filename_component_sep,
                               filename_exclude_str=filename_exclude_str,
                               **kwargs)
        # else:
        #     logger.warn(f'Skipping creating a new FileSerializer object since '
        #                 f'key=(filename_prefix={filename_prefix}, filename_dt_component_type={filename_dt_component_type}, protocol={protocol}, filename_component_sep={filename_component_sep}) '
        #                 f'has been registered!')
        #
        return FileSerializerManager.__registry[key]

    @classmethod
    def get_by_name(cls, name, log_error=True):
        if name not in FileSerializerManager.__named_registry:
            if log_error:
                logger.error(f'Failed to find FileSerializer with name={name}! Please register your FileSerializer first!')

            return None

        return FileSerializerManager.__named_registry[name]

    @classmethod
    def register(cls, name, filename_prefix, filename_dt_component_type,
                 protocol='PKL.GZ',
                 filename_component_sep='.',
                 filename_exclude_str=None,
                 **kwargs):
        if name in FileSerializerManager.__named_registry:
            logger.error(f'name={name} exists in FileSerializerManager.__named_registry! Please choose another unique name!')
            return None

        FileSerializerManager.__named_registry[name] = FileSerializerManager.get(filename_prefix=filename_prefix,
                                                                                 filename_dt_component_type=filename_dt_component_type,
                                                                                 protocol=protocol,
                                                                                 filename_component_sep=filename_component_sep,
                                                                                 filename_exclude_str=filename_exclude_str
                                                                                 **kwargs)
        return FileSerializerManager.__named_registry[name]



import numbers
from collections import OrderedDict
from typing import Iterable
import pandas as pd
from qtc.consts.enums import DollarDenominator, DecimalUnit


def iterable_to_tuple(items, raw_type, sep=',', remove_duplicates=True):
    """Converts "iterable" items into a tuple.

    :param items: Items to be converted. If str, have to be comma separated.
    :type items: str | numbers.Integral | Iterable
    :param raw_type: The data type in the result tuple. 'str' or 'int' are supported.
    :type raw_type: str
    :param sep: Seperator to split items if items is a str.
    :type sep: str
    :param remove_duplicates: To only keep unique items.
    :type remove_duplicates: bool
    :return: tuple

    >>> import qtc.utils.miscs_utils as mu
    >>> mu.iterable_to_tuple('5, 3,-1,3,5,9', raw_type='int')
    (5, 3, -1, 9)
    >>> mu.iterable_to_tuple('5, 3,-1,3,5,9', raw_type='int', remove_duplicates=False)
    (5, 3, -1, 3, 5, 9)
    >>> mu.iterable_to_tuple('Hello', raw_type='str')
    ('Hello',)
    >>> mu.iterable_to_tuple(pd.Series([4,-2,9]), raw_type='str')
    ('4', '-2', '9')
    """
    if isinstance(items, str):
        items = (item.strip() for item in items.split(sep))
        if remove_duplicates:
            items = OrderedDict.fromkeys(items).keys()
    elif isinstance(items, numbers.Integral):
        items = (items, )
    # elif not isinstance(items, list) and not isinstance(items, tuple) \
    #         and not isinstance(items, set) and not isinstance(items, pd.Series):
    elif not isinstance(items, Iterable):
        raise NotImplementedError("The type of parameter 'items' can be only str | numbers.Integral | Iterable")

    raw_type = raw_type.lower()
    if raw_type == 'int':
        items = tuple(int(item) for item in items)
    elif raw_type == 'str':
        items = tuple(str(item) for item in items)

    return items


def preprocess_iterable(items, raw_type, sep=',', remove_duplicates=True):
    if items is None:
        return None

    items = set(iterable_to_tuple(items, raw_type=raw_type, sep=sep, remove_duplicates=remove_duplicates))
    return None if len(items)==0 else items


def iterable_to_db_str(items, raw_type, remove_duplicates=True):
    """This function converts "iterable" items into a str with SQL syntax.

    :param items: Items to be converted. If str, have to be comma separated.
    :type items: str | int | Iterable
    :param raw_type: The data type indicated in the result str. 'str' or 'int' are supported.
    :type raw_type: str
    :param remove_duplicates: To only keep unique items.
    :type remove_duplicates: bool
    :return: str - Object with SQL syntax.

    >>> import qtc.utils.miscs_utils as mu
    >>> mu.iterable_to_db_str('5, 3,-1,3,5,9', raw_type='int')
    '(5,3,-1,9)'
    >>> mu.iterable_to_db_str('5, 3,-1,3,5,9', raw_type='int', remove_duplicates=False)
    '(5,3,-1,3,5,9)'
    >>> mu.iterable_to_db_str('Hello', raw_type='str')
    "('Hello')"
    >>> mu.iterable_to_db_str(pd.Series([4,-2,9]), raw_type='str')
    "('4','-2','9')"
    >>> mu.iterable_to_db_str('Hello, world', raw_type='str')
    "('Hello','world')"
    """
    items = iterable_to_tuple(items, raw_type, remove_duplicates=remove_duplicates)

    if len(items) == 0:
        raise Exception("Parameter 'items' should contain at least one element")

    # items = f"({items[0]})" if len(items) == 1 else str(items)
    if raw_type == 'int':
        items = f"({','.join((str(item) for item in items))})"
    else:
        items = "','".join(items)
        items = f"('{items}')"

    return items


def apply_dollar_denominator_on_df_cols(df, dollar_denominator, cols=None, inplace=False):
    if cols is None:
        return df

    if not inplace:
        df = df.copy()

    if isinstance(dollar_denominator, str):
        dollar_denominator = DollarDenominator.retrieve(value=dollar_denominator)

    for col in iterable_to_tuple(cols, raw_type='str'):
        if col in df.columns:
            if dollar_denominator==DollarDenominator.MILLION:
                df[col] /= 1e6

    return df


def apply_decimal_unit_on_df_cols(df, decimal_unit, cols=None, inplace=False):
    if cols is None:
        return df

    if not inplace:
        df = df.copy()

    if isinstance(decimal_unit, str):
        decimal_unit = DecimalUnit.retrieve(value=decimal_unit)

    for col in iterable_to_tuple(cols, raw_type='str'):
        if col in df.columns:
            if decimal_unit == DecimalUnit.PERCENTAGE:
                df[col] *= 1e+2
            elif decimal_unit == DecimalUnit.BPS:
                df[col] *= 1e+4
            elif decimal_unit != DecimalUnit.DECIMAL:
                raise Exception(f'decimal_unit={decimal_unit} is not supported in [DECIMAL | PERCENTAGE | BPS] !')

    return df


def post_normalize_df(df,
                      dollar_cols=None, decimal_cols=None,
                      dollar_denominator='MILLION', decimal_unit='DECIMAL',
                      inplace=False):
    if not inplace:
        df = df.copy()

    df = apply_dollar_denominator_on_df_cols(df=df, cols=dollar_cols,
                                             dollar_denominator=dollar_denominator, inplace=True)

    df = apply_decimal_unit_on_df_cols(df=df, cols=decimal_cols,
                                       decimal_unit=decimal_unit, inplace=True)

    return df


def standardize_col_names(cols):
    replace_map = {'Exposure': 'Exp',
                   'Variance': 'Var',
                   'Gmv': 'GMV',
                   'Nmv': 'NMV',
                   'Mv': 'MV',
                   'PNL': 'PnL',
                   'PL': 'PnL',
                   'AllocatedCapitalUnit': 'ACU'}

    def standardize_col_name(col):
        for k, v in replace_map.items():
            col = col.replace(k, v)

        return col

    return [standardize_col_name(col) for col in iterable_to_tuple(cols, raw_type='str')]


def get_zero_centered_min_max(values):
    if isinstance(values, list) or isinstance(values, tuple) or isinstance(values, set) or isinstance(values, pd.Series):
        values = list(values)
        min_val = min(values)
        max_val = max(values)
    elif isinstance(values, pd.DataFrame):
        min_val = values.min().min()
        max_val = values.max().max()

    tmp = max(abs(min_val), abs(max_val))
    min_val, max_val = -tmp, tmp
    if min_val == max_val:
        min_val = min_val - 0.0001
        max_val = max_val + 0.0001

    return min_val, max_val


import os
import getpass
import typing as T
from configparser import ConfigParser
from qtc.ext.enum import Enum
from qtc.ext.logging import set_logger
logger = set_logger()


_DEFAULT_ENV = 'RESEARCH'


class Prop(Enum):
    """Enumeration of vanilla property keys with no special handling."""

    FMS_BASE_URL = 'fms.base_url'


class EnvConfig:
    """
    EnvironmentConfig is a singleton that holds runtim environment configuration (instead of application configuration).
    EnvironmentConfig should be treated as an immutable object.

    EnvironmentConfig should be accessed using the qtc.env_config::get_env_config() method.
    For research, it is sometimes useful to pass environment overrides to get_env_config().

    See additional documentation at qtc.env_config::get_env_config()
    """

    _instance: T.Optional['EnvConfig'] = None
    _env_config_file = None

    @classmethod
    def set_env_config_file(cls, env_config_file):
        EnvConfig._env_config_file = env_config_file

    def __init__(self,
                 environment: str,
                 overrides: T.Dict[str, str],
                 show_splash=False,
                 ) -> None:
        self._environment = environment
        self._overrides = overrides
        self._parser = ConfigParser()

        if self._env_config_file is None:
            self._env_config_file = os.path.join(os.path.dirname(__file__), 'env_config.cfg')
        self._parser.read(self._env_config_file)

        if environment not in self._parser.sections():
            raise Exception(f"Unknown environment '{environment}', "
                            f"expected one of {', '.join(self._parser.sections())}")

        if (environment=='PROD') and overrides:
            raise Exception('overrides are not supported in the PROD environment.')

        user = getpass.getuser()
        # if platform.system() == 'Windows':
        #     # user = os.environ['username']
        #     temp_dir = os.environ['tmp']
        # else:
        #     # user = os.environ['USER']
        #     temp_dir = os.environ.get('TMPDIR', f"/tmp/{user}")
        #
        # # self._options holds values of variables that can be used in the
        # # env_config.cfg file.
        # self._options = dict(user=user,
        #                      temp_dir=temp_dir,
        #                      environment=environment)
        self._options = dict(user=user,
                             environment=environment)

        if show_splash:
            logger.info(75 * '#')
            logger.info(f'###  Initializing EnvConfig with env={environment}, user={user}.')
            for key, val in overrides.items():
                logger.info(f'###     [override]   {key} => {val}')
            logger.info(75 * '#')

    @property
    def environment(self) -> str:
        """
        The short environment string that is explicitly passed in with
        the 'SYSEQUITY_ENV' variable or inferred.
        """
        return self._environment

    def get(self, prop, default: T.Optional[str] = None) -> str:
        """
        Gets the value of a property for this environment. If a value is not available, it will use
        a value associated with the [DEFAULT] section in env_config.cfg.
        """
        if isinstance(prop, str):
            try:
                prop = Prop.retrieve(prop).value
            except:
                pass
        if prop in self._overrides:
            return self._overrides[prop]
        try:
            return self._parser.get(self.environment, prop, vars=self._options)
        except:
            return default

    @staticmethod
    def get_instance(env: T.Optional[str] = None,
                     overrides: T.Optional[T.Dict[str, str]] = None,
                     show_splash=False
                     ) -> 'EnvConfig':
        """
        Gets the Singleton instance of the EnvConfig object. Optionally takes an environment
        and a dictionary of property overrides.

        If a environment is not specified, it will default to the 'SYSEQUITY_ENV' environment variable.

        .. note::
            Normally this method is not called directly, instead, please use
            'qtc.env_config::get_env_config' function.
        """
        if EnvConfig._instance is not None:
            if (env is not None) or overrides:
                raise Exception("Only the first call to 'EnvConfig.get_instance()' can use arguments.")
            return EnvConfig._instance

        if env is None:
            env = os.environ.get('CONNECTION_ENV', _DEFAULT_ENV)

        if env is None:
            raise Exception(f'env={env}: The impossible happened!')

        env = env.upper()
        if overrides is None:
            overrides = dict()

        EnvConfig._instance = EnvConfig(env, overrides, show_splash=show_splash)
        return EnvConfig._instance


def get_env_config(env=None, env_config_file=None, show_splash=False, **overrides):
    if env_config_file is None:
        env_config_file = os.path.join(os.path.dirname(__file__), 'env_config.cfg')
    EnvConfig.set_env_config_file(env_config_file=env_config_file)

    return EnvConfig.get_instance(env=env, overrides=overrides, show_splash=show_splash)


import enum
import pandas as pd


class Enum(enum.Enum):
    """An Enum class which allows Enum instance retrievals by Enum item names and/or values."""

    @classmethod
    def retrieve_by_name(cls, value):
        if value is None:
            return None

        for name, item in cls.__members__.items():
            if name.upper() == value.upper():
                return item

        raise ValueError('Failed to find Enum with value {}'.format(value))

    @classmethod
    def retrieve(cls, value):
        if value is None:
            return None

        for name, item in cls.__members__.items():
            if name.upper()==value.upper():
                return item

            if item.value.upper()==value.upper():
                return item

        raise ValueError('Failed to find Enum with value {}'.format(value))

    @classmethod
    def get_meta(cls):
        meta = dict()
        for name, item in cls.__members__.items():
            meta[name] = item.value

        meta = pd.Series(meta).to_frame(name='EnumValue')
        meta.index.name = 'EnumName'
        return meta


import pandas as pd
from datetime import datetime
import pytz
import qtc.utils.datetime_utils as dtu
import qtc.ext.unittest as ut


class TestDateTimeUtils(ut.TestCase):
    def test_datetime_to_dateid(self):
        self.assertEqual(dtu.datetime_to_dateid(datetime(2020, 11, 25)), 20201125)
        adt = datetime(2020, 8, 7, 2, 0, 0, tzinfo=pytz.UTC)
        self.assertEqual(dtu.datetime_to_dateid(adt), 20200807)
        self.assertEqual(dtu.datetime_to_dateid(adt.astimezone(pytz.timezone('America/New_York'))), 20200806)

    def test_dateid_to_datetime(self):
        self.assertEqual(dtu.dateid_to_datetime(20190925), datetime(2019, 9, 25, 0, 0))
        self.assertEqual(dtu.dateid_to_datetime(20190925, timezone='America/New_York'), datetime(2019, 9, 25, 0, 0, tzinfo=pytz.timezone('America/New_York')))

    def test_dateid_to_datestr(self):
        self.assertEqual(dtu.dateid_to_datestr(20201124), '2020-11-24')
        self.assertEqual(dtu.dateid_to_datestr(20201124, sep='/'), '2020/11/24')

    def test_today(self):
        self.assertEqual(dtu.curr_dateid(), dtu.datetime_to_dateid(datetime.now()))
        self.assertEqual(dtu.curr_dateid(timezone='Asia/Tokyo'), dtu.datetime_to_dateid(datetime.now(pytz.timezone('Asia/Tokyo'))))
        self.assertEqual(dtu.curr_dateid(timezone='Europe/London'), dtu.datetime_to_dateid(datetime.now(pytz.timezone('Europe/London'))))
        self.assertEqual(dtu.curr_dateid(timezone='America/New_York'), dtu.datetime_to_dateid(datetime.now(pytz.timezone('America/New_York'))))

    def test_is_weekday(self):
        self.assertEqual(dtu.is_weekday(20200511), True)
        self.assertEqual(dtu.is_weekday(20201128), False)

    def test_get_biz_dates(self):
        self.assertEqual(dtu.get_biz_dateids(start_date=20201120, end_date=20201129), [20201120, 20201123, 20201124, 20201125, 20201126, 20201127])

    def test_next_biz_date(self):
        self.assertEqual(dtu.next_biz_dateid(dateid=20201125), 20201126)
        self.assertEqual(dtu.next_biz_dateid(dateid=20201127), 20201130)

    def test_parse_datetime(self):
        self.assertEqual(dtu.parse_datetime(20200210), datetime(2020, 2, 10, 0, 0))
        self.assertEqual(dtu.parse_datetime('2020-11-25', ret_as_timestamp=True), pd.Timestamp('2020-11-25 00:00:00'))

        target = dtu.parse_datetime('20200210-164523', timezone='America/New_York')
        benchmark = datetime(2020, 2, 10, 16, 45, 23, tzinfo=pytz.timezone('America/New_York'))
        self.assertEqual(target.strftime('%Y%m%d-%H%M%d'), benchmark.strftime('%Y%m%d-%H%M%d'))

        target = dtu.parse_datetime('2020/02/10 16:45:23', format='%Y/%m/%d %H:%M:%S', timezone='Asia/Shanghai')
        benchmark = datetime(2020, 2, 10, 16, 45, 23, tzinfo=pytz.timezone('Asia/Shanghai'))
        self.assertEqual(target.strftime('%Y%m%d-%H%M%d'), benchmark.strftime('%Y%m%d-%H%M%d'))

    def test_normalize_date_to_dateid(self):
        self.assertEqual(dtu.normalize_date_to_dateid(date=20210929), 20210929)
        self.assertEqual(dtu.normalize_date_to_dateid(date='2021-09-29'), 20210929)
        self.assertEqual(dtu.normalize_date_to_dateid(datetime(2021,9,29)), 20210929)

    def test_convert_data_type_for_date_col(self):
        benchmark = self.load_benchmark(basename='convert_data_type_for_date_col.param1.csv',
                                        parse_dates=['Date'])
        data = dtu.convert_data_type_for_date_col(pd.DataFrame({'DateId':[20210927,20210928,20210929]}), date_col='DateId')
        pd.testing.assert_frame_equal(data.head(), benchmark)

    def test_get_week_start_dateid(self):
        self.assertEqual(dtu.get_week_start_dateid(date=20220526), 20220523)
        self.assertEqual(dtu.get_week_start_dateid(date=20220522), 20220516)
        self.assertEqual(dtu.get_week_start_dateid(date=20220523), 20220523)


if __name__ == '__main__':
    ut.main()



import io
import base64
import matplotlib.pyplot as plt


def fig_to_base64(fig=None):
    if fig is None:
        fig = plt.gcf()

    img = io.BytesIO()
    fig.savefig(img, format='png',
                bbox_inches='tight')
    img.seek(0)

    return base64.b64encode(img.getvalue())


def get_default_styles():
    th_col_heading_props = [
        ('text-align', 'center'),
        ('background-color', '#F5F5F5'),
        ('border-collapse', 'collapse'),
        ('border', '1px black solid'),
        ('padding', '5px')
    ]

    th_props = [
        ('text-align', 'center'),
        ('font-size', '12px'),
        ('font-family', 'Arial'),
        # ('background-color', 'white'),
        ('border-collapse', 'collapse'),
        ('border', '1px black solid'),
        ('padding', '5px')
    ]

    td_props = [
        ('text-align', 'right'),
        ('font-size', '12px'),
        ('font-family', 'Arial'),
        ('border-collapse', 'collapse'),
        ('border', '1px black solid'),
        ('padding-left', '5px'),
        ('padding-right', '5px'),
        ('padding-top', '2px'),
        ('padding-bottom', '2px'),
    ]

    styles = [
        dict(selector='th.col_heading', props=th_col_heading_props),
        dict(selector='th', props=th_props),
        dict(selector='td', props=td_props),
        dict(selector='tr:nth-child(even)', props=[('background-color', '#F5F5F5')]),
    ]

    return styles


def html_from_styler(styler, styles=None, title=''):
    if styles is None:
        styles = get_default_styles()

    table_html = styler.set_table_styles(styles)\
                       .set_table_attributes('style="border-collapse:collapse; border:1px black solid"')\
                       .render()

    html = f'<h5>{title}</h5>\n{table_html}' if len(title) else table_html

    return html


def html_from_array(tables, title='', footer=''):
    if len(title):
        title = f'''<h4>{title}</h4>\n'''

    if len(footer):
        footer = f'''<p><i>{footer}</i>'''

    tables = '\n'.join(tables)
    html = f'''
        <html>
            <head>
                <link rel="stylesheet" type="text/css" href="..\limitstyle.css">
            </head>
            <body>
                {title}
                {tables}
                {footer}
            </body>
        </html>
    '''

    return html


def create_td_style_border_text(row_borders, col_borders, border_width='1.0pt'):
    if (row_borders == 'None') & (col_borders == 'None'):
        border_text = ''
    else:
        border_text = f'border:solid windowtext {border_width};'
        if row_borders == 'None':
            border_text = border_text + 'border-top:none;border-bottom:none;'
        elif row_borders == 'Top':
            border_text = border_text + 'border-bottom:none;'
        elif row_borders == 'Bottom':
            border_text = border_text + 'border-top:none;'
        if col_borders == 'None':
            border_text = border_text + 'border-left:none;border-right:none;'
        elif col_borders == 'Left':
            border_text = border_text + 'border-rightnone;'
        elif col_borders == 'Right':
            border_text = border_text + 'border-left:none;'

    return border_text


def create_td_element(width, row_borders, col_borders, data_value=None, height=None, width_pt=None,
                      border_width='1.0pt',
                      align=None, valign=None, nowrap=False, bold_font=False, font_color=None, background=None,
                      colspan=None):
    td_str = f'''<td width={width}'''
    if nowrap:
        td_str = td_str + ' nowrap'
    if valign is not None:
        td_str = td_str + f''' valign={valign}'''
    if colspan is not None:
        td_str = td_str + f''' colspan={colspan}'''

    td_style_str = ''
    if width_pt is not None:
        td_style_str = f'width:{width_pt};'
    td_style_str = td_style_str + create_td_style_border_text(row_borders, col_borders, border_width)
    if background is not None:
        td_style_str = td_style_str + f'''background:{background};'''
    td_style_str = td_style_str + 'padding:0in 5.4pt 0in 5.4pt;'
    if height is not None:
        td_style_str = td_style_str + f'''height:{height};'''
    if len(td_style_str) > 0:
        td_style_str = f" style='{td_style_str}'"

    td_str = td_str + td_style_str + '>'
    if data_value is not None:
        td_str = td_str + '<p class=MsoNormal'
        if align == 'Center':
            td_str = td_str + " style='text-align:center'"
        td_str = td_str + ">"
        if bold_font:
            td_str = td_str + "<b>"
        close_span = False
        if font_color is not None:
            td_str = td_str + f"<span style='color:{font_color}'>"
            close_span = True
        td_str = td_str + str(data_value)
        if close_span:
            td_str = td_str + "</span>"
        # td_str = td_str + "<o:p></o:p>"
        if bold_font:
            td_str = td_str + "</b>"
        td_str = td_str + "</p>"

    td_str = td_str + "</td>"
    return td_str


def format_df_rows(df,
                   formatter,
                   slicer=None):
    df = df.copy()

    if not slicer:
        for col in df.columns:
            df[col] = df[col].apply(formatter.format, axis=0)
        return df
    else:
        dfs = df.loc[slicer].copy()
        for col in dfs.columns:
            dfs[col] = dfs[col].apply(formatter.format, axis=0)
        df.loc[slicer] = dfs
        return df


DEFAULT_UL_BULLET_POINT_CODES = ['&#9632;', '&#9633;', '&#9679;', '&#9900;']

class HTMLListNode:
    def __init__(self, html=''):
        self.html = html
        self.children = list()

    def add_item(self, html):
        node = HTMLListNode(html=html)
        self.children.append(node)
        return node


class HTMLListBuilder:
    def __init__(self,
                 indent_start=0.25, indent_step=0.25,
                 padding=1,
                 ul_bullet_point_codes=DEFAULT_UL_BULLET_POINT_CODES):
        self.root = HTMLListNode()
        self.indent_start = indent_start
        self.indent_step = indent_step
        self.padding = padding
        self.ul_bullet_point_codes = ul_bullet_point_codes

    def _build_html_from_node(self,
                              html_list_node,
                              level=0):
        children = html_list_node.children

        lines = list()
        indent = '\t' * level
        # line = f'{indent}<p class=MsoNormal style="margin-left:{self.indent_start+self.indent_step*level}in;">{html_list_node.html}'
        # lines.append(line)
        for i, child in enumerate(children):
            # indent1 = '\t' * (level+1)
            # line = f'{indent1}<li class="MsoNormal">{child.html}</li>'
            # lines.append(line)
            # number_str = f'({convert_number(number=i+1, level=level)})'
            bullet_point_str = self.ul_bullet_point_codes[level - 1] if level <= len(self.ul_bullet_point_codes) else ''
            p_class = 'MsoListParagraph' if level == 1 else 'MsoNormal'
            line = f'{indent}<p class={p_class} style="margin-left:{self.indent_start + self.indent_step * level}in;">{bullet_point_str}&emsp;{child.html}</p>'
            lines.append(line)

            nested_lines = self._build_html_from_node(html_list_node=child,
                                                      level=level + 1)
            lines.extend(nested_lines)

        return lines

    def build_html_list(self):
        lines = self._build_html_from_node(html_list_node=self.root,
                                           level=1)
        return '\n'.join(lines)


import getpass
import itertools
import qtc.env_config as ecfg
import qtc.utils.endpoint_utils as epu
import typing as T

_BASE_SALT = 'QuantTrading'


def intersperse(str1: str, str2: str) -> str:
    """Mixes together two strings.  If one string is longer, the shorter sting is cycled.
    >>> import sysequity_common.utils.cipher_utils as cu
    >>> cu.intersperse('foo', 'bar')
    'fboaor'
    >>> cu.intersperse('steven', '_-')
    's_t-e_v-e_n-'
    """
    piece_len = max(len(str1), len(str2))
    combined = [None] * 2 * piece_len
    # Mypy doesn't understand that that strings behave as iterable things?  If
    # we let this type float to the 'intersperse' signature, the rest of the
    # file gets ugly.
    combined[0::2] = itertools.islice(itertools.cycle(str1), piece_len) # type: ignore
    combined[1::2] = itertools.islice(itertools.cycle(str2), piece_len) # type: ignore
    return ''.join(map(str, combined))


def scramble(secret: str, *, salt: str = 'AbCdEfG') -> bytes:
    """Apply a salt to a secret string using xor.
    >>> import sysequity_common.utils.cipher_utils as cu
    >>> str(cu.scramble('ABC', salt='ABC'))
    "b'\\\\x00\\\\x00\\\\x00'"
    >>> str(cu.scramble('ABCABC', salt='ABC'))
    "b'\\\\x00\\\\x00\\\\x00\\\\x00\\\\x00\\\\x00'"
    >>> str(cu.scramble('AAA', salt='ABC'))
    "b'\\\\x00\\\\x03\\\\x02'"
    >>> str(cu.scramble('AAAAAA', salt='ABC'))
    "b'\\\\x00\\\\x03\\\\x02\\\\x00\\\\x03\\\\x02'"
    """
    return bytes((ord(a) ^ ord(b)) for (a, b) in zip(secret, itertools.cycle(salt)))


def to_salted(text: str, user: str = None) -> str:
    """Encode a text string by salt
    >>> import sysequity_common.utils.cipher_utils as cu
    >>> cu.to_salted(text='Hello World')
    '0a040d0403552e0e130417'
    """
    if user is None:
        user = getpass.getuser()
    salt = intersperse(_BASE_SALT, user)
    secret = scramble(text, salt=salt)
    secret_str = ''.join('{:02x}'.format(b) for b in secret)
    return secret_str


def from_salted(secret_str: str, user: str = None) -> str:
    """Decodes a salted string
    >>> import sysequity_common.utils.cipher_utils as cu
    >>> cu.from_salted(secret_str='0a040d0403552e0e130417')
    'Hello World'
    """
    if user is None:
        user = getpass.getuser()
    salt = intersperse(_BASE_SALT, user)
    secret_chars = bytearray.fromhex(secret_str)
    secret = ''.join([chr(b) for b in secret_chars])
    return ''.join([chr(b) for b in scramble(secret, salt=salt)])


def get_token():
    url = ecfg.get_env_config().get('token_service.base_url')
    return epu.process_request(request_url=url, is_post=False,
                               data_transmission_protocol='TXT')



from setuptools import setup, find_packages
import os

with open(os.path.join(os.path.dirname(__file__), 'qtc/version.py'), 'r') as f:
    exec(f.read())


setup(
    name='qtc',
    version=__version__,
    url='',
    license='',
    author='Andrew Hu',
    author_email='AndrewWeiHu@gmail.com',
    description='Quant Trading Common',
    packages=find_packages(exclude=['backup']),
    zip_safe=False,
    include_package_data=True,
    platforms='any',
    install_requires=[
        'pandas>=1.3.4',
        'joblib',
        'sqlalchemy',
        'psycopg2',
        'pyarrow',
        'datatable',
        'pytest-html',
        'pyyaml',
        'requests-kerberos',
    ],
    tests_require=[
        'pytest'
    ],
)



from joblib import Parallel, delayed
import qtc.utils.misc_utils as mu
import pandas as pd
import platform


def infer_joblib_backend(use_default=False):
    if use_default:
        return 'loky'

    os_type = platform.system()
    backend = 'threading' if 'Windows' in os_type else 'multiprocessing'
    # backend = 'multiprocessing'
    return backend


def run_multi_dateids_joblib(dateids, func,
                             concat_axis=None,
                             n_jobs=1,
                             **kwargs):
    dateids = mu.iterable_to_tuple(dateids, raw_type='int')
    backend = infer_joblib_backend()
    ret = Parallel(n_jobs=n_jobs, backend=backend)(delayed(function=func)(dateid=dateid, **kwargs)
                                                   for dateid in dateids)
    if concat_axis is not None:
        ret = pd.concat(ret, axis=concat_axis, sort=False)

    return ret


def run_multi_dts_joblib(dts, func,
                         concat_axis=None,
                         n_jobs=1,
                         **kwargs):
    dts = mu.iterable_to_tuple(dts, raw_type='int')
    backend = infer_joblib_backend()
    ret = Parallel(n_jobs=n_jobs, backend=backend)(delayed(function=func)(dt=dt, **kwargs)
                                                   for dt in dts)

    if concat_axis is not None:
        ret = pd.concat(ret, axis=concat_axis, sort=False)

    return ret



import os
import pandas as pd
from joblib import Parallel, delayed
import qtc.utils.datetime_utils as dtu
import qtc.utils.misc_utils as mu
from qtc.consts.enums import DateDataType
from qtc.ext.multiprocessing import infer_joblib_backend
from qtc.file.file_serialization import \
    DateTimeFolderStructureMode, FilenameDateTimeComponentType, FileSerializationProtocol, \
    DateTimeFoldersBuilder, FileSerializer, DateTimeFoldersBuilderManager, FileSerializerManager
from qtc.ext.logging import set_logger
logger = set_logger()


class FileManager:
    @classmethod
    def register(cls):
        if not hasattr(cls, 'NAME'):
            logger.warn(f"Skipping since no class variable 'NAME' in {cls} !")
            return None

        if cls.NAME is None:
            logger.warn(f"Skipping since {cls}.NAME is None !")

        if not hasattr(cls, 'DATETIME_FOLDERS_BUILDER'):
            logger.warn(f"Skipping since no class variable 'DATETIME_FOLDERS_BUILDER' in {cls} !")
            return None

        if cls.DATETIME_FOLDERS_BUILDER is None or not isinstance(cls.DATETIME_FOLDERS_BUILDER, DateTimeFoldersBuilder):
            logger.warn(f"Skipping since {cls}.DATETIME_FOLDER_BUILDER is None or not of type DateTimeFoldersBuilder !")

        file_serializer_dict = None
        if hasattr(cls, 'FILE_SERIALIZER_DICT'):
            file_serializer_dict = cls.FILE_SERIALIZER_DICT

        FileManagerRegistry.register(name=cls.NAME,
                                     file_manager_cls=cls,
                                     datetime_folders_builder=cls.DATETIME_FOLDERS_BUILDER,
                                     file_serializer_dict=file_serializer_dict)

    @property
    def datetime_folders_builder(self):
        return self.__datetime_folders_builder

    @datetime_folders_builder.setter
    def datetime_folders_builder(self, datetime_folders_builder):
        self.__datetime_folders_builder = datetime_folders_builder

    @property
    def file_serializer_dict(self):
        return self.__file_serializer_dict

    @file_serializer_dict.setter
    def file_serializer_dict(self, file_serializer_dict):
        self.__file_serializer_dict = file_serializer_dict

    @property
    def root(self):
        return self.datetime_folders_builder.root

    @property
    def datetime_folder_structure_mode(self):
        return self.datetime_folders_builder.datetime_folder_structure_mode

    # def __init__(self, name,
    #              root, datetime_folder_structure_mode,
    #              filename_prefix, filename_dt_component_type,
    #              protocol='PKL.GZ',
    #              filename_component_sep='.',
    #              timezone=None,
    #              **kwargs):
    #     self.datetime_folders_builder = None
    #     self.file_serializers = list()
    # #

    def __init__(self, datetime_folders_builder, file_serializer_dict=None):
        if datetime_folders_builder is not None:
            if not isinstance(datetime_folders_builder, DateTimeFoldersBuilder):
                raise Exception(f'datetime_folders_builder={datetime_folders_builder} is not of type DateTimeFoldersBuilder !')

        self.datetime_folders_builder = datetime_folders_builder

        self.file_serializer_dict = dict()
        if file_serializer_dict is not None:
            if not isinstance(file_serializer_dict, dict):
                raise Exception(f'file_serializer_dict={file_serializer_dict} has to be of type dict(key:str, value:DateTimeFoldersBuilder) !')

            for name, file_serializer in file_serializer_dict.items():
                if not isinstance(file_serializer, FileSerializer):
                    logger.warn(f'With name={name}, file_serializer={file_serializer} is not of type FileSerializer!')
                    continue

                self.file_serializer_dict[name] = file_serializer

    def get_file_serializer(self, file_serializer):
        if isinstance(file_serializer, str):
            if file_serializer not in self.file_serializer_dict:
                file_serializer_from_registry = FileSerializerManager.get_by_name(name=file_serializer, log_error=False)
                if file_serializer_from_registry is None:
                    logger.error(f'No file_serializer with name={file_serializer} neither in this class object nor global FileSerializerManager!')
                    return None
                else:
                    return file_serializer_from_registry
            else:
                return self.file_serializer_dict[file_serializer]
        elif isinstance(file_serializer, FileSerializer):
            return file_serializer

        logger.warn(f'file_serializer={file_serializer} is not a str nor a FileSerializer !')
        return None

    def __repr__(self):
        class_variables_repr = '\n\t'
        datetime_folders_builder_class = None
        file_serializer_dict_class = None
        # print(type(self))
        if hasattr(type(self), 'NAME'):
            class_variables_repr += f'NAME: {type(self).NAME}' + '\n\t'
        if hasattr(type(self), 'DATETIME_FOLDERS_BUILDER'):
            datetime_folders_builder_class = type(self).DATETIME_FOLDERS_BUILDER
            class_variables_repr += f'DATETIME_FOLDERS_BUILDER: {datetime_folders_builder_class}' + '\n\t'
        if hasattr(type(self), 'FILE_SERIALIZER_DICT'):
            file_serializer_dict_class =  type(self).FILE_SERIALIZER_DICT
            class_variables_repr += f'FILE_SERIALIZER_DICT: {file_serializer_dict_class}' + '\n\t'

        repr = f'{type(self).__name__}({class_variables_repr}' \
               f'\n\tdatetime_folders_builder: {self.datetime_folders_builder if self.datetime_folders_builder!=datetime_folders_builder_class else "DATETIME_FOLDERS_BUILDER"}'\
               f'\n\tfile_serializer_dict: {self.file_serializer_dict if self.file_serializer_dict!=file_serializer_dict_class else "FILE_SERIALIZER_DICT"}'\
               f'\n)'
        return repr

    def compile_folder(self, dt, subfolders=None, is_subfolders_under_dt_folder=True):
        folder = self.datetime_folders_builder.compile_folder(dt=dt, subfolders=subfolders,
                                                              is_subfolders_under_dt_folder=is_subfolders_under_dt_folder)
        return folder

    def compile_file_path(self, file_serializer,
                          dt=None, subfolders=None, is_subfolders_under_dt_folder=True,
                          extra_filename_components=None,
                          include_run_dt=False):

        file_serializer = self.get_file_serializer(file_serializer=file_serializer)

        if file_serializer is None:
            raise Exception(f'Failed to obtain a FileSerializer instance!')

        folder = self.compile_folder(dt=dt, subfolders=subfolders,
                                     is_subfolders_under_dt_folder=is_subfolders_under_dt_folder)

        filename = file_serializer.compile_filename(dt=dt,
                                                    extra_filename_components=extra_filename_components,
                                                    timezone=self.datetime_folders_builder.timezone,
                                                    include_run_dt=include_run_dt)
        return os.path.join(folder, filename)

    def serialize(self, data,
                  file_serializer,
                  dt=None, subfolders=None, is_subfolders_under_dt_folder=True,
                  extra_filename_components=None,
                  include_run_dt=False):
        file_serializer = self.get_file_serializer(file_serializer=file_serializer)

        if file_serializer is None:
            raise Exception(f'Failed to obtain a FileSerializer instance!')

        file_path = self.compile_file_path(file_serializer=file_serializer,
                                           dt=dt, subfolders=subfolders, is_subfolders_under_dt_folder=is_subfolders_under_dt_folder,
                                           extra_filename_components=extra_filename_components,
                                           include_run_dt=include_run_dt)
        file_path = file_serializer._serialize(data=data, file_path=file_path)
        return file_path

    def search_file(self, file_serializer,
                    dt=None, subfolders=None, is_subfolders_under_dt_folder=True,
                    extra_filename_components=None,
                    allow_as_of=False,
                    log_error=True):
        file_serializer = self.get_file_serializer(file_serializer=file_serializer)

        if file_serializer is None:
            raise Exception(f'Failed to obtain a FileSerializer instance!')

        folder = self.compile_folder(dt=dt, subfolders=subfolders,
                                     is_subfolders_under_dt_folder=is_subfolders_under_dt_folder)
        filename = file_serializer.search_file(folder=folder,
                                               dt=dt,
                                               extra_filename_components=extra_filename_components,
                                               timezone=self.datetime_folders_builder.timezone, allow_as_of=allow_as_of,
                                               log_error=log_error)
        if filename is None:
            return None

        return os.path.join(folder, filename)

    def _deserialize(self, file_serializer,
                     dt=None, subfolders=None, is_subfolders_under_dt_folder=True,
                     extra_filename_components=None,
                     allow_as_of=False,
                     dt_col=None,
                     log_error=True,
                     ret_filepath=False,
                     **kwargs):
        file_serializer = self.get_file_serializer(file_serializer=file_serializer)

        file_path = kwargs.get('file_path', None)
        if file_path is None:
            file_path = self.search_file(file_serializer=file_serializer,
                                         dt=dt, subfolders=subfolders, is_subfolders_under_dt_folder=is_subfolders_under_dt_folder,
                                         extra_filename_components=extra_filename_components,
                                         allow_as_of=allow_as_of,
                                         log_error=log_error)

        data = file_serializer._deserialize(file_path=file_path, log_error=log_error)

        try:
            if dt_col is not None and dt_col not in data.columns:
                data[dt_col] = dtu.normalize_dt(dt=dt, timezone=self.datetime_folders_builder.timezone)
        except:
            pass

        return (data, file_path) if ret_filepath else data

    def deserialize(self, file_serializer,
                    start_dt=None, end_dt=None, dts=None,
                    subfolders=None,
                    extra_filename_components=None,
                    allow_as_of=False,
                    dt_type='DateId',
                    log_error=False,
                    to_df=True,
                    **kwargs):

        if isinstance(dt_type, str):
            dt_type = DateDataType.retrieve(dt_type)

        if dts is None:
            if dt_type == DateDataType.DATEID:
                dts = list(dtu.get_biz_dateids(start_date=start_dt, end_date=end_dt))
        else:
            dts = list(mu.iterable_to_tuple(dts, raw_type='str'))
            if dt_type==DateDataType.DATEID:
                dts = [dtu.datetime_to_dateid(dtu.normalize_dt(dt, timezone=self.datetime_folders_builder.timezone))
                       for dt in dts]

        dt_col = 'DateTime'
        file_serializer = self.get_file_serializer(file_serializer=file_serializer)
        backend = infer_joblib_backend()
        ret = Parallel(n_jobs=-1, backend=backend)(delayed(unwrap_FileManager_self__deserialize)\
                                                           (i,
                                                            file_serializer=file_serializer,
                                                            subfolders=subfolders,
                                                            extra_filename_components=extra_filename_components,
                                                            allow_as_of=allow_as_of,
                                                            dt_col=dt_col,
                                                            log_error=log_error,
                                                            **kwargs)
                                                   for i in zip([self]*len(dts), dts))
        ret = [item for item in ret if item is not None]
        if len(ret)==0:
            return None

        if to_df:
            ret = pd.concat(ret, ignore_index=True, sort=False)

            if dt_type==DateDataType.DATEID:
                ret = dtu.convert_data_type_for_date_col(df=ret, date_col=dt_col,
                                                         from_data_type=DateDataType.TIMESTAMP,
                                                         to_data_type=DateDataType.DATEID,
                                                         to_date_col_idx=None)
                ret.drop(columns=[dt_col], inplace=True)

        return ret


### Have to do it in this ugly way coz joblib is not able to handle locals in class obj
def unwrap_FileManager_self__deserialize(arg, **kwarg):
    return FileManager._deserialize(arg[0], **dict(kwarg, dt=arg[1]))


class FileManagerRegistry:
    __registry = dict()

    @classmethod
    def register(cls, name, file_manager_cls, datetime_folders_builder, file_serializer_dict=None):
        if name not in FileManagerRegistry.__registry:
            FileManagerRegistry.__registry[name] = file_manager_cls(datetime_folders_builder=datetime_folders_builder,
                                                                    file_serializer_dict=file_serializer_dict)
        else:
            logger.error(f'NAME has been registered for:\n{FileManagerRegistry.__registry[name]}')

        return FileManagerRegistry.__registry[name]

    # @classmethod
    # def initialize(cls):
    #     for subclass in FileManager.__subclasses__():
    #         logger.debug(f'Registering subclass: {subclass}')
    #         subclass.register()
    #         logger.debug(f'subclass: {subclass} successfully registered.')
    #     #
    # #

    @classmethod
    def get(cls, name, root=None):
        if name not in FileManagerRegistry.__registry:
            raise Exception(f'name={name} not found in FileManagerRegistry.registry !')

        file_manager = FileManagerRegistry.__registry[name]
        if root is None or root==file_manager.datetime_folders_builder.root:
            return file_manager

        datetime_folders_builder = DateTimeFoldersBuilderManager.get(root=root,
                                                                     datetime_folder_structure_mode=file_manager.datetime_folders_builder.datetime_folder_structure_mode,
                                                                     timezone=file_manager.datetime_folders_builder.timezone)
        file_manager = type(file_manager)(datetime_folders_builder=datetime_folders_builder,
                                          file_serializer_dict=file_manager.file_serializer_dict)
        return file_manager


def get_meta():
    datetime_folder_structure_mode_enums = DateTimeFolderStructureMode.get_meta()
    datetime_folder_structure_mode_enums['EnumClass'] = 'DateTimeFolderStructureMode'
    filename_dt_component_type_enums = FilenameDateTimeComponentType.get_meta()
    filename_dt_component_type_enums['EnumClass'] = 'FilenameDateTimeComponentType'
    protocol_enums = FileSerializationProtocol.get_meta()
    protocol_enums['EnumClass'] = 'FileSerializationProtocol'

    meta = pd.concat([datetime_folder_structure_mode_enums,
                      filename_dt_component_type_enums,
                      protocol_enums])
    meta.index.name = 'EnumName'
    return meta.reset_index().set_index(['EnumClass','EnumName'])



import unittest
import qtc.utils.cipher_utils as cu


class TestCipherUtils(unittest.TestCase):
    def test_intersperse(self):
        self.assertEqual(cu.intersperse('foo', 'bar'), 'fboaor')
        self.assertEqual(cu.intersperse('steven', '_-'), 's_t-e_v-e_n-')

    def test_to_salted(self):
        self.assertEqual(cu.to_salted(text='Hello World'), '1910190e0e553901061830')

    def test_from_salted(self):
        self.assertEqual(cu.from_salted(secret_str='1910190e0e553901061830'), 'Hello World')


if __name__ == '__main__':
    unittest.main()




from abc import ABC
from bq_configs.enums import *
from dataclasses import dataclass, fields
from math import inf
from typing import Optional, _GenericAlias, get_args, get_origin  # type: ignore


@dataclass
class TradeRules:
    min_trade_dollar: float = 0.0  #: filters out all trades with $ value below min_trade_size
    max_trade_dollar: float = inf  #: $ amounts
    max_trade_adv: float = inf  #: percentage of ADV
    min_trade_risk: float = 0  #: if (trades * V * trades’)^0.5 < min_trade_risk then no trade
    

@dataclass
class OrderManagerParamsBase(ABC):
    trade_rules: Optional[TradeRules] = None  #: trade rules
    fill_type: FillType = FillType.Close  #: fill type
    pov_cap: float = 0.1  #: 


@dataclass
class PositionManagerParamsBase(ABC):
    drift_pos: bool = True  #: Account for position drift
    

@dataclass
class PortfolioOptimizationParamsBase(ABC):
    strategy: str = 'BAM-Example-TEST'  #: Strategy name
    
    optimizer_type: OptimizerType = OptimizerType.BAMOptimizer  #: For optimizer_type: 'bam_optimizer' for the cloud optimizer, "MOSEK" for the MOSEK solver through cvxpy
    opt_url: Optional[str] = None  #: url to cloud-based optimizer. Will try to use getenv('BAM_OPTIMIZER_URL') if explicit value is not provided
    
    # GMV
    target_gmv: float = 250e6  #: Target Gross Market Value
    nmv_as_fraction_gmv: float = inf  #: Net Market Value as fraction of GMV, ex.: 0.1 
    
    # Transaction Cost
    tc_spread_scale: float = 0  #: Transaction Cost, ex.: 0.000001
    market_impact_exponent: float = 1.0  #: FIXME need some validation for this number
    tc_scale: float = 0.0  #: Transaction Cost scale, if set to 0 then no transaction cost
    srisk_power: float = 0.0  #: 

    # Positions
    turnover_limit: float = inf  #: Cap on the sum of the absolute value of all trades as percentage of GMV, can be inf, ex.: 0.2
    position_fraction_adtv: float = inf  #: Cap for each position as percentage of ADV, ex.: 0.01 for 1%
    trade_fraction_adtv: float = inf  # Cap for each trade as percentage of ADV, ex.: 0.01 for 1%
    position_fraction_gmv: float = inf  #: Cap for each position as percentage of GMV, ex.: 0.1 for 10%

    # Risk
    factor_risk_penalty: float = 0  #: ex.: 3.0e-3
    srisk_penalty: float = 0  #: ex.: 1.0e-8
    srisk_floor: float = 0  # default to zero, not needed.

    # Alpha
    penalty_scale_distance_from_ideal: float = 0  #: Set to 0 if return optimization; set to non-zero for deviation from ideal optimization

    # Volatility
    portfolio_vol: float = 0  #: default to zero, not needed.
    upper_total_factor_vol: Optional[float] = None  #: 
    upper_total_idio_vol: Optional[float] = None  #: 
    upper_total_vol: Optional[float] = None  #: 

    # Other
    # TODO used in OptimizerInput only. Potential duplicated by scale down
    scale_factor: float = 1  #: unused, see below scale_down (rescale the problem when x is either very large or very small)
    post_opt_check: bool = True  #: For non BAMOptimizer optimizers, check if fulfilled the constraints within a threshold
    min_srisk_pct: float = 0  #: 80 means 80%, Ensures specific risk makes up at least this much percent of total risk. Set to 0 to not enable
    # trade_rules: dict = field(default_factory=lambda: {'no_trade_dow': 'Mon Tue Thr', 'no_trade_cal': None})  #: Unused
    scale_down: float = 1e6  #: Scale down factor. A typical value is 1e6, which scales down by that factor. This is an important tool to use if you find the run is crashing
    auto_scale: bool = True  #: Automatically set scale_down based on GMV
    mkt_impact_scale_factor: float = 1  #: If set to 0 then no Market Impact Transaction Cost

    signal_value_type: SignalValueType = SignalValueType.ExpectedReturns  #: 
    default_ba: float = 0.0015  #: The default value for bid ask in intraday
    
    backtest: bool = True  #: 

    custom_factors_src_path: str = None

    # Additional constraints
    pos_bounds_cols: Optional[tuple] = ('Id', 'PositionLB1', 'PositionUB1')  #: columns for DataFrame of additional position constraints
    pos_bounds_data: Optional[list] = None  #: $ amounts, e.g. [['BBG000B9XRY4', -1000000, 5000000]]
    factor_constraints_cols: Optional[tuple] = ('FactorId', 'Upper', 'Lower')  #: columns for DataFrame of additional factor constraints
    factor_constraints_data: Optional[list] = None  #: [["Market Sensitivity", 10000.0, -10000.0], ["Volatility", 10000.0, -50000.0], ["Exchange Rate Sensitivity", 20000.0, -20000.0]]
    factor_constraint_type: str = 'dollar'  # options: 'dollar', 'percent'. Percent if % target_gmv

def dict2dataclass(cls, dict_args: dict):
    typed_dict = {}
    for f in fields(cls):
        if not f.init:
            continue
        
        value = dict_args.get(f.name)
        if value is None:
            continue
        if get_origin(f.type):
            if isinstance(value, dict):
                typed_dict[f.name] = get_args(f.type)[0](**value)
            else:
                typed_dict[f.name] = get_args(f.type)[0](value)
        else:
            typed_dict[f.name] = f.type(value)

    return cls(**typed_dict)



def dataclass2dict(cls) -> dict:
    res = {}
    for f in fields(cls):
        if not f.init:
            continue
        if isinstance(getattr(cls, f.name), TradeRules):
            res[f.name] = to_dict(getattr(cls, f.name))
        elif issubclass(type(f.type), _GenericAlias):
            res[f.name] = getattr(cls, f.name)
        elif issubclass(f.type, Enum):
            res[f.name] = getattr(cls, f.name).value
        else:
            res[f.name] = getattr(cls, f.name)

    return res


def to_dict(arg) -> dict:
    return {f.name: getattr(arg, f.name) for f in fields(arg)}




import bisect
import re
from dateutil.relativedelta import relativedelta
import qtc.utils.datetime_utils as dtu
import qtc.utils.misc_utils as mu
from qtc.consts.enums import OffsetDateType
from qtc.ext.logging import set_logger
logger = set_logger()

# class CalendarType(Enum):
#     EXCHANGE = 'Exchange'
#     FACTOR_MODEL = 'FactorModel'


class Calendar:
    DEFAULT_TIME_ZONE = 'America/New_York'

    @staticmethod
    def _get_trading_dateids(trading_dateids, start_dateid=None, end_dateid=None):
        start_dateid = min(trading_dateids) if start_dateid is None else int(start_dateid)
        end_dateid = max(trading_dateids) if end_dateid is None else int(end_dateid)

        return (dateid for dateid in trading_dateids if dateid >= start_dateid and dateid <= end_dateid)

    @staticmethod
    def is_trading_date(dateid, trading_dateids):
        return dateid in trading_dateids

    @staticmethod
    def prev_trading_dateid(dateid, trading_dateids):
        if dateid <= trading_dateids[0]:
            logger.error(f'dateid={dateid} is on or earlier than the first dateid {trading_dateids[0]} !')
            return None

        i = bisect.bisect_left(trading_dateids, dateid)
        return trading_dateids[i - 1] if trading_dateids[i] in trading_dateids else trading_dateids[i]

    @staticmethod
    def next_trading_dateid(dateid, trading_dateids):
        if dateid >= trading_dateids[-1]:
            logger.error(f'dateid={dateid} is on or later than the last dateid {trading_dateids[-1]} !')
            return None

        i = bisect.bisect_right(trading_dateids, dateid)
        return trading_dateids[i]

    @staticmethod
    def shift_trading_days(dateid, trading_dateids, offset):
        if offset is None or not isinstance(offset, int):
            logger.warn(f"Skipping since 'offset' is not properly provided!")
            return dateid
        #
        shift_func = Calendar.prev_trading_dateid if offset <= 0 else Calendar.next_trading_dateid
        offset = abs(offset)

        for i in range(offset):
            dateid = shift_func(dateid=dateid, trading_dateids=trading_dateids)

        return dateid

    @staticmethod
    def get_asof_trading_dateid(trading_dateids, dateid=None,
                                timezone=DEFAULT_TIME_ZONE):
        if dateid is None:
            dateid = dtu.curr_dateid(timezone=timezone)

        return dateid if Calendar.is_trading_date(dateid=dateid, trading_dateids=trading_dateids) else \
               Calendar.prev_trading_dateid(dateid=dateid, trading_dateids=trading_dateids)

    @staticmethod
    def get_trading_dateids(trading_dateids, start_date=None, end_date=None, dates=None):
        if dates is None:
            if start_date is None or end_date is None:
                raise Exception(f'Please provide either "dates" or ("start_date", "end_date")')

            start_dateid = dtu.normalize_date_to_dateid(date=start_date)
            end_dateid = dtu.normalize_date_to_dateid(date=end_date)

            dateids = Calendar._get_trading_dateids(start_dateid=start_dateid, end_dateid=end_dateid,
                                                    trading_dateids=trading_dateids)
        else:
            dates = mu.iterable_to_tuple(dates, raw_type='int')
            dateids = list()
            for date in dates:
                dateid = dtu.normalize_date_to_dateid(date=date)
                if dateid in trading_dateids:
                    dateids.append(dateid)

        return list(dateids)

    @staticmethod
    def infer_offset_dateid(ref_date, offset_date, trading_dateids):
        ref_date = dtu.parse_datetime(dt=ref_date)

        if isinstance(offset_date, str):
            ndays_matched = re.match(r"([-+]?\d+)D$", offset_date, re.IGNORECASE)
            if ndays_matched is not None:
                ndays = int(ndays_matched.groups()[0])
                offset_date = Calendar.shift_trading_days(dateid=dtu.normalize_date_to_dateid(date=ref_date),
                                                          trading_dateids=trading_dateids,
                                                          offset=ndays)
            else:
                offset_date = OffsetDateType.retrieve(value=offset_date)

                if offset_date == OffsetDateType.CURR_MONTH_END:
                    offset_date = dtu.get_month_end(date=ref_date)
                elif offset_date == OffsetDateType.NEXT_MONTH_END:
                    date = ref_date.replace(year=ref_date.year + 1, month=1, day=1) if ref_date.month == 12 else \
                           ref_date.replace(month=ref_date.month + 1, day=1)
                    offset_date = dtu.get_month_end(date=date)
                else:
                    raise Exception(f'future_date={offset_date} not supported in [CME|NME|*D] !')

        offset_dateid = dtu.normalize_date_to_dateid(date=offset_date)
        if not Calendar.is_trading_date(dateid=offset_dateid, trading_dateids=trading_dateids):
            offset_dateid = Calendar.next_trading_dateid(dateid=offset_dateid, trading_dateids=trading_dateids)

        return offset_dateid

    @staticmethod
    def infer_start_dateid_end_dateid(trading_dateids,
                                      start_date=None, end_date=None, date_range_mode='SINGLE_DATE',
                                      default_ctd=False, timezone=DEFAULT_TIME_ZONE):

        curr_dateid = dtu.curr_dateid(timezone=timezone)
        if end_date is None:
            end_dateid = curr_dateid
            if not default_ctd:
                end_dateid = Calendar.prev_trading_dateid(dateid=end_dateid, trading_dateids=trading_dateids)
        else:
            if end_date == 'PTD':
                end_dateid = Calendar.prev_trading_dateid(dateid=curr_dateid, trading_dateids=trading_dateids)
            elif end_date == 'CTD':
                end_dateid = curr_dateid
            else:
                end_dateid = dtu.normalize_date_to_dateid(date=end_date)
                if not Calendar.is_trading_date(dateid=end_dateid, trading_dateids=trading_dateids):
                    end_dateid = Calendar.prev_trading_dateid(dateid=end_dateid, trading_dateids=trading_dateids)

        if start_date is not None:
            start_dateid = dtu.normalize_date_to_dateid(date=start_date)
            if not Calendar.is_trading_date(dateid=start_dateid, trading_dateids=trading_dateids):
                start_dateid = Calendar.next_trading_dateid(dateid=start_dateid, trading_dateids=trading_dateids)
            if start_dateid <= end_dateid:
                return start_dateid, end_dateid
            else:
                logger.warn(f"Inconsistency found: start_date={start_dateid} > end_date={end_dateid}. "
                            f"start_date will be inferred by date_range_mode={date_range_mode} instead.")

        ndays_matched = re.match(r"(\d+)D$", date_range_mode, re.IGNORECASE)
        nweeks_matched = re.match(r"(\d+)W$", date_range_mode, re.IGNORECASE)
        nmonths_matched = re.match(r"(\d+)M$", date_range_mode, re.IGNORECASE)

        if date_range_mode == 'SINGLE_DATE':
            start_dateid = end_dateid
        elif date_range_mode == 'ROLLING_WEEK':
            start_dateid = Calendar.shift_trading_days(dateid=end_dateid, trading_dateids=trading_dateids, offset=-4)
        elif date_range_mode == 'ROLLING_MONTH':
            start_date = dtu.dateid_to_datetime(dateid=end_dateid) - relativedelta(months=1)
            start_dateid = dtu.datetime_to_dateid(date=start_date)
            if not Calendar.is_trading_date(dateid=start_dateid, trading_dateids=trading_dateids):
                start_dateid = Calendar.next_trading_dateid(dateid=start_dateid, trading_dateids=trading_dateids)
        elif date_range_mode == 'MTD':
            start_dateid = int(end_dateid / 100.0) * 100 + 1
            if not Calendar.is_trading_date(dateid=start_dateid, trading_dateids=trading_dateids):
                start_dateid = Calendar.next_trading_dateid(dateid=start_dateid, trading_dateids=trading_dateids)
        elif ndays_matched is not None:
            ndays = int(ndays_matched.groups()[0])
            start_dateid = Calendar.shift_trading_days(dateid=end_dateid, trading_dateids=trading_dateids, offset=-(ndays-1))
        elif nweeks_matched is not None:
            ndays = int(nweeks_matched.groups()[0]) * 5
            start_dateid = Calendar.shift_trading_days(dateid=end_dateid, trading_dateids=trading_dateids, offset=-(ndays - 1))
        elif nmonths_matched is not None:
            nmonths = int(nmonths_matched.groups()[0])
            start_date = dtu.dateid_to_datetime(dateid=end_dateid) - relativedelta(months=nmonths)
            start_dateid = dtu.datetime_to_dateid(date=start_date)
            if not Calendar.is_trading_date(dateid=start_dateid, trading_dateids=trading_dateids):
                start_dateid = Calendar.next_trading_dateid(dateid=start_dateid, trading_dateids=trading_dateids)
        else:
            raise Exception(f"Supported 'date_range_mode' are [SINGLE_DATE|ROLLING_WEEK|ROLLING_MONTH|MTD|$(ndays)D|$(nweeks)W|$(nmonths)M] !")

        return start_dateid, end_dateid

    @staticmethod
    def infer_trading_dateids(trading_dateids,
                              start_date=None, end_date=None, date_range_mode='SINGLE_DATE'):
        # start_dateid = None if start_date is None else dtu.normalize_date_to_dateid(date=start_date)
        # end_dateid = None if end_date is None else dtu.normalize_date_to_dateid(date=end_date)

        start_dateid, end_dateid = \
            Calendar.infer_start_dateid_end_dateid(trading_dateids=trading_dateids,
                                                   start_date=start_date, end_date=end_date,
                                                   date_range_mode=date_range_mode)
        dateids = list(Calendar.get_trading_dateids(trading_dateids=trading_dateids,
                                                    start_date=start_dateid, end_date=end_dateid))

        return dateids, start_dateid, end_dateid